#!/usr/bin/env python
# gloTK - Genomes of Luminous Organisms Toolkit
# Copyright (c) 2015-2016 Darrin Schultz. All rights reserved.
#
# This file is part of gloTK.
#
# GloTK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GloTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GloTK.  If not, see <http://www.gnu.org/licenses/>.

#import things for rest of program
import sys
import argparse
import os
import re
import subprocess
import shutil

#https://py-gfm.readthedocs.io/en/latest/
import markdown
from mdx_gfm import GithubFlavoredMarkdownExtension


class MerRunAnalyzer:
    """This class generates a report for one meraculous run. It requires a run
    directory to look for the report, as well as a directory to deposit the
    report. The generate_report() method will output an HTML file and folder
    in the report directory. It will use the meraculous run name as the name for
    the report HTML file and report directory.
    """
    def __init__(self, run_directory, output_parent_dir,
                 censor, quiet=False):
        """In a future edit, change this method to auto-detect whether the
        run is diploid mode 1, 2, or 3
        """

        self.home = os.path.abspath(run_directory.strip())
        self.reportName = os.path.basename(run_directory.strip())
        self.parent = os.path.dirname(self.home.strip())

        #setup the bit to cleanly censor the files
        self.home.strip().rsplit("/")
        censor = [x for x in self.parent.strip().rsplit("/") if x]
        self.censor = {strip: "" for strip in censor}
        self.rep = dict((re.escape(k), v) for k, v in self.censor.items())
        self.pattern = re.compile("|".join(self.rep.keys()))

        #filename prefix, sanitized with censor method
        self.outname=self._str_ripper(os.path.basename(self.home))
        self.merReportsDir = os.path.join(self.parent, "reports")
        self.reportDir = os.path.join(self.merReportsDir, self.outname)

        #make sure that the directory is a meraculous directory
        self.isMer = self._is_meraculous()
        self.quiet = quiet

        self.mer_size = None
        self.diploid_mode = None
        self.genome_size = None
        self.min_depth_cutoff = None
        if self.isMer:
            self._get_Params()

    def _is_meraculous(self):
        subdirs = [dirs.strip() for dirs in os.listdir(self.home) if os.path.isdir(os.path.join(self.home,dirs.strip()))]
        return ("log" in subdirs) and ("meraculous_import" in subdirs)

    def _call_sys(self, callString, logfile):
        print("\ncalling command:", file=logfile)
        print(self._str_ripper(callString), file=logfile)
        p = subprocess.run(callString, shell=True, stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE,
                           universal_newlines=True)
        output = self._str_ripper(str(p.stdout))
        err = self._str_ripper(str(p.stderr))
        print(output, file=logfile)
        print(err, file=logfile)
        # if not self.quiet:
        #     print(output)
        #     print(err)
        return (output, err)

    def _make_HTML(self, logfile):
        html_filepath = os.path.join(
            self.merReportsDir,
            "{0}_report.html".format(self.outname))
        print("making html at {}".format(html_filepath))
        callString = "python -m markdown -x markdown.extensions.fenced_code -x pymdownx.b64 {0} > {1}".format(
            logfile,
            html_filepath)
        p = subprocess.run(callString, shell=True)

    def _get_Params(self):
        logFile = os.path.join(self.home, "log/meraculous.log")
        grep_list=["mer_size", "diploid_mode",
                   "genome_size", "min_depth_cutoff"]
        done = {x:False for x in grep_list}

        with open(logFile, "r") as f:
            for line in f:
                for query in grep_list:
                    if re.search(query, line):
                        value=line.strip().split()[1]
                        #print("query: {}, value: {}".format(query, value))
                        setattr(self, query, value)
                        done[query] = True
                if False not in done.values():
                    break

    def _str_ripper(self, text):
        """Got this code from here:
        http://stackoverflow.com/questions/6116978/python-replace-multiple-strings

        This method takes a set of strings, A, and removes all whole
        elements of set A from string B.

        Input: text string to strip based on instance attribute self.censor
        Output: a stripped (censored) text string
        """
        return self.pattern.sub(lambda m: self.rep[re.escape(m.group(0))], text)

    def generate_report(self):
        """This method does most of the work for this class. It generates an
        individual report for the instance.
        """

        print("making HTML report for {}".format(self.outname))
        # make the markdown file to log to while doing analyses
        logfile_filepath = os.path.join(
            self.reportDir,
            "{0}_report.md".format(self.outname))

        #check if seld.reportDir exists
        # if not, make it
        if not os.path.exists(self.reportDir):
            print("making the directory: {}".format(self.reportDir))
            os.makedirs(self.reportDir)

        logfile = open(logfile_filepath, 'w')

        # make a title in the markdown file
        print("# {0}".format(self.outname), file=logfile)
        print("## Meraculous assembly QC analysis", file=logfile)

        # 0. Run params
        print("### 0. Run params", file=logfile)
        print("", file=logfile)
        print("- mer_size: `{}`".format(self.mer_size), file=logfile)
        print("- diploid_mode: `{}`".format(self.diploid_mode), file=logfile)
        print("- genome_size: `{}`".format(self.genome_size), file=logfile)
        print("- min_depth_cutoff: `{}`".format(self.min_depth_cutoff), file=logfile)
        print("", file=logfile)

        # 1. get the mercount and kha plot
        print("### 1. `meraculous_mercount` output", file = logfile)
        print("", file=logfile)
        print("#### `mercount.png`", file = logfile)
        print("", file=logfile)
        mercountFromPath = os.path.join(self.home, "meraculous_mercount/mercount.png")
        mercountToPath = os.path.join(self.reportDir, "mercount.png")
        mercountHTML = os.path.join(
            os.path.basename(os.path.split(mercountToPath)[0]),
            os.path.basename(mercountToPath))
        shutil.copyfile(mercountFromPath, mercountToPath)
        print("![mercount]({0})".format(mercountHTML),
            file=logfile)
        print("", file=logfile)
        print("#### `kha.png`", file = logfile)
        print("", file=logfile)
        khaFromPath = os.path.join(self.home, "meraculous_mercount/kha.png")
        khaToPath = os.path.join(self.reportDir, "kha.png")
        khaHTML = os.path.join(
            os.path.basename(os.path.split(khaToPath)[0]),
            os.path.basename(khaToPath))
        shutil.copyfile(khaFromPath, khaToPath)
        print("![kha]({0})".format(khaHTML),
            file=logfile)
        print("", file=logfile)

        # 2. run fasta_stats on UUtig.fa
        # this is the result of the first contigs produced by meraculous_contigs
        print("### 2. `fasta_stats UUtigs.fa`", file=logfile)
        print("", file=logfile)
        print("""This is the output of `meraculous_contigs`. In a diploid genome,
              the total length of all contigs at this stage should be
              larger than the expected genome size because of the
              presence of haplotype variant UUtigs. (Meraculous
              manual)""", file=logfile)
        print("",file=logfile)
        print("```", file=logfile)
        callString="fasta_stats {0}".format(
            "{}/meraculous_contigs/UUtigs.fa".format(self.home))
        self._call_sys(callString, logfile)
        print("```", file=logfile)
        print("",file=logfile)

        if int(self.diploid_mode) == 1:
            #2.5. Check out the output of meraculous_bubble
            print("### 2.5. Diploid mode plot", file=logfile)
            print("", file=logfile)
            print("""- If there are two peaks, and the "half-depth" peak is much larger
                  than the one at full depth, that's a scenario that's
                  best handled by diploid_mode 2.  If it's the
                  opposite, make sure the parameter
                  'bubble_depth_threshold' is being set correctly (if
                  auto-detected, the result is written into
                  `meraculous_bubble/haplotigs.dmin.err`) - E. Goltsman.""",
                  file=logfile)
            print("""- For diploid assemblies, you should examine the file
                  haplotigs.depth.hist.png and verify that there are
                  two distinct peaks, one at roughly half depth of the
                  other. At this point you may want to check your
                  bubble_depth_threshold parameter and adjust it to a
                  value corresponding to the local minimum between the
                  two peaks (_ if you had originally set it to 0
                  Meraculous auto-detects this threshold_). - Meraculous Manual""",
                  file=logfile)
            print("", file=logfile)
            bblFromPath = os.path.join(self.home,
                                       "meraculous_bubble/haplotigs.depth.hist.png")
            bblToPath = os.path.join(self.reportDir, "haplotigs.depth.hist.png")
            bblHTML = os.path.join(
                os.path.basename(os.path.split(bblToPath)[0]),
                os.path.basename(bblToPath))
            shutil.copyfile(bblFromPath, bblToPath)
            print("![bbl]({0})".format(bblHTML),
                   file=logfile)
            print("", file=logfile)
            bbl_detect=os.path.join(self.home, "meraculous_bubble/haplotigs.dmin.err")
            if os.path.exists(bbl_detect):
                #in a later version, just look at the meraculous log file
                print("", file=logfile)
                print("#### `bubble_depth_threshold` auto-detected", file=logfile)
                print("", file=logfile)
                print("```", file=logfile)
                callString="cat {0}".format(bbl_detect)
                self._call_sys(callString, logfile)
                print("```", file=logfile)
                print("", file=logfile)
            else:
                print("", file=logfile)
                print("#### `bubble_depth_threshold` was user-input", file=logfile)
                print("", file=logfile)
            print("#### `fasta_stats meraculous_bubble/haplotigs.fa", file=logfile)
            print("", file=logfile)
            print("""- These are the assembly stats after running
                  `meraculous_bubble`.""", file=logfile)
            print("", file=logfile)
            print("```", file=logfile)
            callString="fasta_stats {0}".format(
            os.path.join(self.home, "meraculous_bubble/haplotigs.fa"))
            self._call_sys(callString, logfile)
            print("```", file=logfile)
            print("", file=logfile)

        if int(self.diploid_mode) == 1:
            print("### 2.6. Note on `meraculous_ono` with `diploid_mode=1`", file=logfile)
            print("", file=logfile)
            print("""- [ Note: If running in diploid_mode 1, the scaffolding is initially
                performed using combined linkage info from alternative
                variant diplotigs, i.e., both variants contribute read
                pairs to the same link as if they were one and the
                same contig. Then, using haplotype-specific read
                mapping info, phased variant paths are determined and
                the scaffold content is corrected in a haplotype
                consistent manner. As a result, one variant path
                (typically one with the higher overall depth) will be
                preserved in a multi-contig scaffold while the
                individual alternative variants are represented as
                unlinked, singleton scaffolds. A list of these
                singleton alternative variant scaffolds is also
                saved. For more on diploid-aware assembly, see section
                ‘Diploid assembly’. ] - _Meraculous manual_""", file=logfile)

        print("### 3. `meraculous_gap_closure`", file=logfile)
        print("", file=logfile)
        print("#### 3.1 `final.scaffolds.fa`", file=logfile)
        print("", file=logfile)
        print("```", file=logfile)
        callString="fasta_stats {0}".format(
            os.path.join(self.home, "meraculous_gap_closure/final.scaffolds.fa"))
        self._call_sys(callString, logfile)
        print("```", file=logfile)
        print("", file=logfile)

        if int(self.diploid_mode) == 1:
            print("#### 3.1. Note on haplotype final scaffolds", file=logfile)
            print("", file=logfile)
            print("""- [Note: when running in diploid_mode 1, if a gap
            represents a polymorphic region that had been actively
            removed earlier (i.e., a half-depth isotig), Meraculous
            will attempt to walk across it using reads from the more
            abundant allele]. - _Meraculous manual_""", file=logfile)
            print("""- When running in diploid_mode 1, a “single-haplotype” sequence
            file (final.scaffolds.single-haplotype.fa) where the
            alternative variant scaffolds have been removed is also
            created. - _Meraculous manual_""", file=logfile)
            print("", file=logfile)
            print("```", file=logfile)
            callString="fasta_stats {0}".format(
                os.path.join(self.home, "meraculous_gap_closure/final.scaffolds.single-haplotype.fa"))
            self._call_sys(callString, logfile)
            print("```", file=logfile)
            print("", file=logfile)

        print("### 4. `meraculous_final_results`", file=logfile)
        print("""- The final output of meraculous""", file=logfile)
        print("", file=logfile)
        print("```", file=logfile)
        callString="fasta_stats {0}".format(
            os.path.join(self.home, "meraculous_final_results/final.scaffolds.fa"))
        self._call_sys(callString, logfile)
        print("```", file=logfile)
        print("", file=logfile)

        # this block handles the printing the meraculous log to the new output
        # directory
        print("### 5. `meraculous_log`", file=logfile)
        print("""- The final output of meraculous""", file=logfile)
        print("", file=logfile)
        print("```", file=logfile)
        callString="cat {0}".format(
            os.path.join(self.home, "log/meraculous.log"))
        output, err = self._call_sys(callString, logfile)
        print("```", file=logfile)
        print("", file=logfile)
        new_mer_log_path = os.path.join(self.reportDir, "meraculous.log")
        with open (new_mer_log_path, "w") as f:
            print(output, file=f)

        logfile.close()
        self._make_HTML(logfile_filepath)




#!/usr/bin/env python3
# gloTK - Genomes of Luminous Organisms Toolkit
# Copyright (c) 2015-2016 Darrin Schultz. All rights reserved.
#
# This file is part of gloTK.
#
# GloTK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GloTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GloTK.  If not, see <http://www.gnu.org/licenses/>.

"""

title: MerParse.py
authr: darrin schultz
vrsin: 0.01

v0.01 - Tue Aug  9 18:28:25 PDT 2016
      - Started file

This program will do a few things:
1. It will read in a meraculous config file and parse the input.
2. It will write new config files from the input and perform a multi-assembly
   parameter sweep. Simply put, it runs many assemblies through a user-defined
   parameter range so that the user can find the optimum assembly.
"""
from .libseq import LibSeq
from collections import UserDict
from numbers import Number
from time import strftime as tfmt
from string import ascii_letters, digits
import os

class MerParse:
    """This class:

    1. reads in an input file and extrapolates all of the possible
       parameters for Meraculous
    2. Choose a parameter to sweep on and calculates the sweep intervals
    3. Output all of the config files to <$PWD>/mer_configs
    4. Return a list of filepaths for the main method to use for assembly and
        report generation.

    Useage example:
    myMerParser = MerParse(<params>)
    myPaths = myMerParser.sweeper_output()

    Now the myPaths object contains a dict of run names and absolute paths for
        the config files.
    """
    def __init__(self, inputFile, sweep, sStart, sStop, sInterval, lnProcs,
                 asPrefix = "as", asSI = 0, genus = None, species = None):
        # --------------- Config File Parameters -------------------------------
        self.inputFile = inputFile
        self.params = {"lib_seq": [],
                       "genome_size": -0.1,
                       "mer_size": -1,
                       "min_depth_cutoff": 0,
                       "num_prefix_blocks": -1,
                       "diploid_mode": 0,
                       "use_cluster": 0,
                       "no_read_validation": 0,
                       "fallback_on_est_insert_size": 0,
                       "gap_close_aggressive": 0,
                       "gap_close_rpt_depth_ratio": 2.0,
                       "local_num_procs": 1,
                       "local_max_retries": 0}
        self.diploid_mode = {"bubble_depth_threshold": 0,
                             "strict_haplotypes": 1}
        # makes two lists that contain which strings refer to things that should
        #  be converted to ints vs floats
        self.to_int = [x for x in self.params if isinstance(self.params[x], int)] + [
            y for y in self.diploid_mode if isinstance(self.diploid_mode[y], int)]
        self.to_float = [x for x in self.params if isinstance(self.params[x], float)] + [
            y for y in self.diploid_mode if isinstance(self.diploid_mode[y], float)]

        # for debugging purposes, makes a list of values that were assigned a
        #  negative value upon initialization, showing that they were not
        #  specified in the config file and need to be changed. Raise an error
        #  in the read_config file.
        # self.config_specified are things that must be specified in the config
        #  file, so if they are negative we know that the user made a mistake.
        # self.config_unspecified are things that have a default value already,
        #  so if they are negative then the user mis-entered something in the
        # config file.
        self.config_specified = [x for x in [u for u in self.params if isinstance(self.params[u], Number)] if self.params[x] < 0]
        self.config_unspecified = [x for x in [u for u in self.params if isinstance(self.params[u], Number)] if self.params[x] >= 0] + [
            y for y in [v for v in self.diploid_mode if isinstance(self.diploid_mode[v], Number)] if self.diploid_mode[y] >= 0]

        # -------------------------Sweep Parameters-----------------------------
        self.sweep = sweep
        self.sweep_support_int = ["mer_size", "bubble_depth_threshold"]
        self.sStart = sStart
        self.sStop = sStop
        self.sInterval = sInterval

            #make sure that they're supported sweep types
        if self.sweep not in self.sweep_support_int:
            raise AttributeError("""{0} is not a currently supported sweep-able
            parameter. Please choose one of the following parameters or update
            the software at https://github.com/cypridina/gloTK. {1}""".format(
                self.sweep, self.sweep_support))

            #if they're supposed to be ints, convert them to ints first
        if self.sweep in self.sweep_support_int:
            for each in ["sStart", "sStop", "sInterval"]:
                setattr(self, each, int(getattr(self, each)))
                attr=getattr(self, each)
                if attr % 1 != 0:
                    raise AttributeError("""One or more of your sweep parameters
                    contains float numbers {0}, but you chose to sweep on a parameter
                    that requires ints {1}. Please try again.""".format(each, self.sweep))

        if self.sStart >= self.sStop:
            raise AttributeError("""The input sweep start value is greater than
            or equal to the sweep stop value. {0}>={1}. Please make sweep start
            less than sweep stop.""".format(self.sStart, self.sStop))
        if ((self.sStop - self.sStart) % self.sInterval) != 0:
            raise AttributeError("""Your sweep interval does not end on your
            sweep stop, and is off by {0}. Try increasing or decreasing your
            sweep stop by that much.""".format(self.sInterval - ((self.sStop - self.sStart) % self.sInterval)))
                #------------------------------Sweep List-------------------------------
        # Generate list for the sweep parameters. Current implementation only
        #   sweeps through ints, so no special processing is needed.
        # If sweep is "mer_size", then make sure that none of the values are
        #   even.
        if self.sweep in self.sweep_support_int:
            sweep_list = [x for x in range(self.sStart, self.sStop + self.sInterval, self.sInterval)]
            if self.sweep == "mer_size":
                evens = [x for x in sweep_list if x % 2 == 0]
                if evens:
                    raise AttributeError("""Your sweep interval for mer_size
                    resulted in the following even numbers: {0}. Please adjust
                    your sweep start, stop, and interval to only contain
                    positive odd integers""".format(evens))
            self.sList = sweep_list

        # -------------------------Name Parameters------------------------------
        self.as_a = asPrefix
        self.as_i = int(asSI)
        #http://stackoverflow.com/questions/311627/
        self.as_d = tfmt("%Y%m%d")
        self.as_z = "ME"
        self.find_illegal_characters(genus, species)
        self.as_g = genus if genus else None
        self.as_s = species if species else None
        # local number of processors (number of processors for one run)
        self.lnProcs = lnProcs

        #-----------------------Directory Parameters----------------------------
        self.cwd = os.getcwd()
        #-----------------------Read the config file----------------------------
        self.read_config()


    def find_illegal_characters(self, genus, species):
        #only allow the user to input ASCII letters and digits (no punctuation)
        # in the genus and species names to avoid filename errors
        illegal = [x for x in
                   str(genus).replace("None","") + str(species).replace("None","")
                   if x not in ascii_letters + digits]
        if illegal:
            raise IOError("""<genus> or <species/sample> contains the
                illegal characters: {}""".format(illegal))

    def name_gen(self, as_number, k):
        """See the sweeper_output() method doctstring for format details."""
        #how to add prefixes to numbers
        #http://stackoverflow.com/a/135157
        #         ai_d_z_g_s_k

        parts =  {"1ai": "{0}{1:03d}_".format(self.as_a, as_number),
                  "2d" : "{0}_".format(self.as_d),
                  "3z" : "ME_",
                  "4g" : "{}_".format(self.as_g) if self.as_g else "",
                  "5s" : "{}_".format(self.as_s) if self.as_s else "",
                  "6k" : "k{}".format(k)}
        return "".join([parts.get(x) for x in sorted(parts)])

    def sweeper_output(self):
        """This method outputs all of the config files in the configs directory
        within the current working directory. It names the config files based on
        the following format:

        a = assembly prefix <default = as>
        i = assembly start index <default = 000>
        d = date in format YYYYMMDD <default=today>
        z = assembler type - ME for Meraculous in this case
        g = first four letters of genus. Truncates if species name is too long.
        s = first four letters of species. Truncates if genus name is too long.
        k = kmer size

        ai_d_z_g_s_k

        For example, if I enter the following parameters for a k=21 run on
        20160811:
          - g = "Malacosteus"
          - s = "niger"

        The assembly directory will be saved in the current working
        directory as:
          - configs/as000_20160811_ME_mala_nige_k21

        Steps for this method:
        1. Check if configs is a directory that exists, if not, make it.
        2. For loop
          - Assign assembly numbers to sweep param
          - Output config file for each sweep param
        3. return dict of {"<run string>": "<config abs path>"}
        """
        # 1. check if configs directory exists, if not, make it
        config_dir = os.path.join(os.getcwd(), "configs")
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)

        # 2. Assign assembly numbers to sweep param
        asNum_sweep_dict = {asNum:self.sList[asNum - self.as_i]
                            for asNum in range(self.as_i, self.as_i + len(self.sList))}
        if self.sweep == "mer_size":
            asName_asSweep_dict = {
                self.name_gen(asNum, asNum_sweep_dict[asNum]): asNum_sweep_dict[asNum]
                                      for asNum in asNum_sweep_dict}
        else:
            asName_asSweep_dict = {
                self.name_gen(asNum, self.params.get("mer_size")): asNum_sweep_dict[asNum]
                                  for asNum in asNum_sweep_dict}

        for name in asName_asSweep_dict:
            sweep = asName_asSweep_dict[name]
            with open(os.path.join(config_dir, name), "w") as f:
                self.params[self.sweep] = sweep
                for key in self.params:
                    if key == "lib_seq":
                        for each in self.params["lib_seq"]:
                            print("lib_seq {0}".format(each), file = f)
                    else:
                        print("{0} {1}".format(key,self.params[key]), file=f)

        # 3 return dict of {"<run string>": "<config abs path>"}
        return {name:os.path.join(config_dir, name)
                for name in asName_asSweep_dict}

    def assign(self, dict_name, key, value):
        """This assigns an input string to either a float or int and saves it in
        the config params as such"""
        if key in self.to_int:
            getattr(self, dict_name)[key] = int(value)
        elif key in self.to_float:
            getattr(self, dict_name)[key] = float(value)
        else:
            raise ValueError("""The param:value pair, {0}:{1}, has neither been
            identified as an int nor a float. Therefore, we are not able to save
            its value to any dictionary in the merParse class""".format(
                key, value))

    def libseq_parse(self, line):
        """This method parses a line containing a lib_seq string and adds it to
        the lib_seq key in the self.params dictionary"""

        LS = LibSeq()
        line = line.split()
        if (len(line) - 1) != 12:
            raise ValueError("""ERROR: There are {0} values in the lib_seq line when
                             there should only be 10. Check your input meraculous
                             config file for errors and make sure that it matches
                             the specification for the Meraculous manual.""".format(
                                 len(line)))
        for index in LS.values:
            LS[LS.values.get(index)] = line[index]
        return LS

    def space_error(self, line):
        raise ValueError(
            """ERROR: There are too many spaces in this line in your config
            file: {0}""".format(line))

    def read_config(self):
        with open(self.inputFile, "r") as f:
            for line in f:
                line=line.strip()
                # if the line is empty, this will prevent observing those lines
                if line:
                    if line[0] is not "#":
                        split = line.split()
                        param = split[0]
                        val = split[1]
                        if param in self.params:
                            if param == "lib_seq":
                                #add a libseq object to the libseq key in self.params
                                self.params["lib_seq"].append(self.libseq_parse(line))
                            else:
                                if len(split) == 2:
                                    self.assign("params", param, val)
                                else:
                                    #make sure there is only a param and value here,
                                    # otherwise there must be an error in the config
                                    # file on this line
                                    self.space_error(line)

                        elif param in self.diploid_mode:
                            if len(split) == 2:
                                self.assign("diploid_mode", param, val)
                            else:
                                # should only be 2 elements long
                               self.space_error(line)
                        # If the param isn't in either of those two dictionaries, it
                        # hasn't been implemented yet
                        else:
                            print("""This line is not legal input for this version of
                            glotk-mer-ksweep.py. Please remove it or consult
                            the software author via
                            https://github.com/cypridina/gloTK

                            {0}""".format(line))
        # check to make sure that the following things are not negative values,
        # signalling that they were never appropriately changed
        for tempDict in [self.params, self.diploid_mode]:
            for key in tempDict:
                val = tempDict.get(key)
                # This avoids comparasion problems with the LibSeq objects
                if key != "lib_seq":
                    if (val < 0) and (key in self.config_specified):
                        #see the __init__() method at the self.config_specified
                        # definition for a better description
                        raise ValueError(
                            """The param {0} was not specified in the config file.
                            Please specify it in the config file and re-run this
                            script""".format(key))
                    elif (val < 0) and (key in self.config_unspecified):
                        raise ValueError(
                            """The param {0} was incorrectly specified as a negative
                            number in the config file. Please use only positive
                            integers or positive floats in the config file and try
                            again.""".format(key))
        # override the number of procs based on parallelism controlled in
        #  gloTK/scripts/glotk_sweep.py
        self.params["local_num_procs"] = self.lnProcs



#!/usr/bin/env python3
# gloTK - Genomes of Luminous Organisms Toolkit
# Copyright (c) 2015-2016 Darrin Schultz. All rights reserved.
#
# This file is part of gloTK.
#
# GloTK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GloTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GloTK.  If not, see <http://www.gnu.org/licenses/>.

"""

title: libseq.py
authr: Darrin Schultz

This is the LibSeq class that is used to store data for the "lib_seq" information
in the Meraculous config files.
"""

from collections import UserDict

class LibSeq(UserDict):
    def __init__(self):
        self.values = {1: "wildcard",
                  2: "name",
                  3: "insertAvg",
                  4: "insertSdev",
                  5: "avgReadLn",
                  6: "hasInnieArtifact",
                  7: "isRevComped",
                  8: "useForContiging",
                  9: "scaffRound",
                  10: "useForGapClosing",
                  11: "5p_wiggleRoom",
                  12: "3p_wiggleRoom"}
        self.data = {self.values.get(x): "" for x in self.values}

    def __repr__(self):
        print_str = ""
        for index in sorted(self.values):
            value = self.data.get(self.values.get(index))
            print_str += "{0} ".format(value.replace(" ",""))
        return print_str[0:-1]



#!/usr/bin/env python3

# Tutorials on how to setup python package here:
#   - http://python-packaging.readthedocs.io/en/latest/testing.html
#   - https://jeffknupp.com/blog/2013/08/16/open-sourcing-a-python-project-the-right-way/

from setuptools import setup

setup(name='gloTK',
      version='0.1',
      description='Genomes of Luminous Organisms Toolkit',
      long_description="""
          GloTK is a toolkit for enhancing productivity in the process
          of whole genome de novo assembly. Currently the package
          supports automated assembly HTML report generation and automated
          assemly parameter optomization for the Meraculous assembler.""",
      url='https://github.com/cypridina/gloTK',
      author='Darrin Schultz',
      author_email='dschultz@mbari.org',
      classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Operating System :: POSIX :: Linux',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Intended Audience :: Science/Research'
          ],
      license='GPLv3',
      provides=['gloTK'],
      packages=['gloTK', 'gloTK.scripts'],
      install_requires=[
          #MerParse - None
          #tests - nose (see below)
          #MerRunAnalyzer
          "py_gfm",
          "pymdown-extensions",
          "markdown"
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      entry_points = {
            'console_scripts': ['glotk-sweep=gloTK.scripts.glotk_sweep:main'],
        },
      zip_safe=False,
      include_package_data=True)



#!/usr/bin/env python
# gloTK - Genomes of Luminous Organisms Toolkit
# Copyright (c) 2015-2016 Darrin Schultz. All rights reserved.
#
# This file is part of gloTK.
#
# GloTK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GloTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GloTK.  If not, see <http://www.gnu.org/licenses/>.

"""

title: glotk-sweep
authr: darrin schultz

This program:
1. Reads in a meraculous config file and outputs all of the associated config
   files to $PWD/configs
2. The name of each run and the path to the directory is passed to a
   multiprocessing core that controls which assemblies are executed and when.
3. Each assembly is executed.

"""

#import things for rest of program
import sys
import argparse
import os

#stuff for call shell commands
import subprocess

#multiprocessing stuff
from functools import partial
from multiprocessing import cpu_count
from multiprocessing import Pool
from multiprocessing.dummy import Pool as ThreadPool

#import gloTK stuff
from gloTK import MerParse

#This class is used in argparse to expand the ~. This avoids errors caused on
# some systems.
class FullPaths(argparse.Action):
    """Expand user- and relative-paths"""
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest,
                os.path.abspath(os.path.expanduser(values)))

class CommandLine:
    """
    authors: David Bernick and Darrin Schultz
    Handle the command line, usage and help requests.
    """
    def __init__(self) :
        """For stage 1, the necessary arguments for the MerParse class are:
          - inputFile
          - sweep
          - sStart
          - sStop
          - sInterval
          - asPrefix
          - asSI
          - genus
          - species
        """
        self.parser=argparse.ArgumentParser(description=__doc__)
        self.parser.add_argument("-i", "--inputConfig",
                            type=str,
                            action=FullPaths,
                            required=True,
                            help="""The meraculous config file upon which the
                            sweep will be based.""")
        self.parser.add_argument("-s", "--sweep",
                            type=str,
                            choices=["mer_size", "bubble_depth_threshold"],
                            required=True,
                            help="""The parameter to sweep through in the
                            assembly.""")
        self.parser.add_argument("--sstart",
                            help="""The start for the sweep parameter""")
        self.parser.add_argument("--sstop",
                            help="""The stop for the sweep parameter""")
        self.parser.add_argument("--sinterval",
                            help="""The interval between the sweep start and stop.""")
        self.parser.add_argument("-p", "--prefix",
                            type=str,
                            help="""The assembly prefix used to name config files
                            and the output directories.""")
        self.parser.add_argument("-I", "--index",
                            type=str,
                            help="""The starting assembly index for naming the config
                            files and output directory.""")
        self.parser.add_argument("-G", "--genus",
                            type=str,
                            help="""The genus name for the sample that will be used
                            for naming config files and directory names.""")
        self.parser.add_argument("-S", "--species",
                            type=str,
                            help="""The species name for the sample that will be used
                            for naming config files and directory names.""")
        self.parser.add_argument("-n", "--simultaneous",
                            type=int,
                            default=1,
                            help="""The number of simultaneous assemblies to run.""")
        self.parser.add_argument("-M", "--maxProcs",
                            type=int,
                            default=cpu_count() - 2,
                            help="""The total number of processers used by all
                            of the assemblies combined.""")
        self.parser.add_argument("-c", "--censor",
                            type=str,
                            nargs='+',
                            help="""list of strings to censor for sending docs to
                            non-collaborators""")
        self.parser.add_argument("-q", "--quiet",
                            action='store_true')


        self.args = self.parser.parse_args()

        # check for input errors
        self._check_errors()

    def _check_errors(self):
        """Call this function after parsing the args to see if there are any
        errors in the way things are input. Specifically for glotk-sweep, make
        sure that all of the parameters for sweep have arguemnts if at least one
        does."""
        # use this to make sure that sweep, sstart, sstop, and sinterval
        # all have values if one of them has values
        # (P^Q^R^S) OR (-P and -Q and -R and -S)
        # above is the minimal form for this logical statement in DNF
        if not ((self._inn(self.args.sweep) and self._inn(self.args.sstart)
                 and self._inn(self.args.sstop) and self._inn(self.args.sinterval))
                or
                (not self._inn(self.args.sweep) and not self._inn(self.args.sstart) and not
                 self._inn(self.args.sstop) and not self._inn(self.args.sinterval))):
            mutually_inclusive = ["sweep", "sstart", "sstop", "sinterval"]
            print_str = ""
            for each in mutually_inclusive:
                print_str += "    {0}: {1}\n".format(each, getattr(self.args, each))
            raise AssertionError("""You specified one or more of the --sweep,
            --sstart, --stop, or --sinterval arguments but did not specify all of
            them. All of them are required when running the program in sweep mode.
            Do not specify any of these arguments if not running the program in
            sweep mode.\n{0}""".format(print_str))

    def _inn(self, value):
        """Checks to verify that an argument is not None, aka that it has a value.
        Lets you use boolean logic on arguments since Python doesn't have
        mutually inclusive groups."""
        return value is not None

def mer_runner_dummy(instance):
    """This method is a helper method for class MerRunner. It allows
    multiprocessing module in python to map parallelism to a class method."""
    instance.meraculous_runner()

class MerRunner:
    """This class has one instance per Meraculous run and is accessed with the
    partial module"""
    def __init__(self, runName, configPath, cleanup=0):
        """The cleanup parameter is what is passed to the run_meraculous script"""
        self.runName = runName
        self.configPath = configPath
        self.cleanup =  cleanup
        self.cwd = os.path.abspath(os.getcwd())
        self.allAssembliesDir = os.path.join(self.cwd, "assemblies")
        self.thisAssemblyDir = os.path.join(self.allAssembliesDir, self.runName)
        self.reportsDir = os.path.join(self.cwd, "reports")

        self.callString = "bash run_meraculous.sh -c {0} -dir {1} -cleanup_level {2}".format(
            self.configPath, self.runName, self.cleanup)

    def meraculous_runner(self):
        """
        Check to make sure that the allAssembliesDir has been created, if not,
        make it. This will only execute for the first time an assembly has been
        run in this directory.

        Run the directory from allAssembliesDir. The self.callString instance
        attribute tells Meraculous to name the assembly directory self.runName.

        After the run is complete, create the meraculous report, passing the
        directory containing the run (aka self.thisAssemblyDir).
        """
        #set the dir to temp assembly dir
        os.chdir(self.allAssembliesDir)

        print(self.callString)
        p = subprocess.run(self.callString, shell=True, stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE,
                           universal_newlines=True)
        output = str(p.stdout)
        err = str(p.stderr)

        #generate the report for the run
        self._generate_report()

        #exit, returning the output and err
        return (output, err)

    def _generate_report(self):
        reporter = MerRunAnalyzer(self.thisAssemblyDir, self.cwd, [])
        reporter.generate_report()

def main():
    """
    1. Reads in a meraculous config file and outputs all of the associated config
       files to $PWD/configs
    2. The name of each run and the path to the directory is passed to a
       multiprocessing core that controls which assemblies are executed and when.

    """
    print(sys.argv)
    if len(sys.argv) <= 1:
        print("""Please input some options for this program or see how it is used.""")
        parser = CommandLine(["--help"])
    else :
        parser = CommandLine()
        myArgs = parser.args
        print(myArgs)

    #Figure out how many processors to give to each assembly since we will be
    # running some things in parallel. The MerParse class will handle overriding
    # whatever is found in the config file in the read_config() method.
    procsPerAssembly = int(myArgs.maxProcs / myArgs.simultaneous)
    setattr(myArgs, "maxProcs", procsPerAssembly)

    # 1. Reads in a meraculous config file and outputs all of the associated config
    #    files to $PWD/configs

    merparser = MerParse(myArgs.inputConfig, myArgs.sweep, myArgs.sstart,
                         myArgs.sstop, myArgs.sinterval, myArgs.maxProcs,
                         asPrefix = myArgs.prefix,
                         asSI = myArgs.index,
                         genus = myArgs.genus,
                         species = myArgs.species)
    configPaths = merparser.sweeper_output()

    #make the assemblies dir ONCE to avoid a race condition for os.makedirs()
    cwd = os.path.abspath(os.getcwd())
    allAssembliesDir = os.path.join(cwd, "assemblies")
    if not os.path.exists(allAssembliesDir):
        os.makedirs(allAssembliesDir)



    #instantiate all of the classes that we will be using in parallel processing.
    # configPaths above returns a dict with the run name and abs path of config
    # as key:value pairs
    instances = []
    for runName in configPaths:
        configPath = configPaths.get(runName)
        thisInstance = MerRunner(runName, configPath)
        instances.append(thisInstance)

    if len(instances) == 0:
        print("There are no meraculous folders in this directory. Exiting")
    elif len(instances) > 0:
        # run the program for each instance
        # pool size is the number of simultaneous runs for the server
        pool = ThreadPool(myArgs.simultaneous)
        results = pool.map(mer_runner_dummy, instances)
        pool.close()
        pool.join()

if __name__ == "__main__":
    sys.exit(main())
